/* check.c */

/*
 * Copyright (c) 1997-2002 Hoshi Takanori
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <PalmOS.h>

#include "life.h"


/*
 * constants
 */

#define GLIDER_LENGTH 5
#define GLIDER_PERIOD 4

#define GLIDER_MARGIN 4

struct glider {
    cell_t adjust[GLIDER_PERIOD][GLIDER_LENGTH];
    cell_t next;
};

static struct glider glider_up_left = {
  {
    { ADJUST(0, 0), ADJUST(1, 0), ADJUST(2, 0), ADJUST(0, 1), ADJUST(1, 2) },
    { ADJUST(1, -1), ADJUST(0, 0), ADJUST(1, 0), ADJUST(0, 1), ADJUST(2, 1) },
    { ADJUST(0, -1), ADJUST(1, -1), ADJUST(0, 0), ADJUST(2, 0), ADJUST(0, 1) },
    { ADJUST(0, -1), ADJUST(1, -1), ADJUST(-1, 0), ADJUST(0, 0), ADJUST(1, 1) }
  },
    ADJUST(-1, -1)
};

static struct glider glider_up_right = {
  {
    { ADJUST(0, 0), ADJUST(1, 0), ADJUST(2, 0), ADJUST(2, 1), ADJUST(1, 2) },
    { ADJUST(1, -1), ADJUST(1, 0), ADJUST(2, 0), ADJUST(0, 1), ADJUST(2, 1) },
    { ADJUST(1, -1), ADJUST(2, -1), ADJUST(0, 0), ADJUST(2, 0), ADJUST(2, 1) },
    { ADJUST(1, -1), ADJUST(2, -1), ADJUST(2, 0), ADJUST(3, 0), ADJUST(1, 1) }
  },
    ADJUST(1, -1)
};

static struct glider glider_down_left = {
  {
    { ADJUST(2, 0), ADJUST(1, 0), ADJUST(0, 0), ADJUST(0, -1), ADJUST(1, -2) },
    { ADJUST(1, 1), ADJUST(1, 0), ADJUST(0, 0), ADJUST(2, -1), ADJUST(0, -1) },
    { ADJUST(1, 1), ADJUST(0, 1), ADJUST(2, 0), ADJUST(0, 0), ADJUST(0, -1) },
    { ADJUST(1, 1), ADJUST(0, 1), ADJUST(0, 0), ADJUST(-1, 0), ADJUST(1, -1) }
  },
    ADJUST(-1, 1)
};

static struct glider glider_down_right = {
  {
    { ADJUST(2, 0), ADJUST(1, 0), ADJUST(0, 0), ADJUST(2, -1), ADJUST(1, -2) },
    { ADJUST(1, 1), ADJUST(2, 0), ADJUST(1, 0), ADJUST(2, -1), ADJUST(0, -1) },
    { ADJUST(2, 1), ADJUST(1, 1), ADJUST(2, 0), ADJUST(0, 0), ADJUST(2, -1) },
    { ADJUST(2, 1), ADJUST(1, 1), ADJUST(3, 0), ADJUST(2, 0), ADJUST(1, -1) }
  },
    ADJUST(1, 1)
};


/*
 * functions
 */

static int is_glider(
    const cell_t *ptr, Boolean up, const struct glider *glider,
    cell_t *origin_ptr)
{
    int i, j;

    for (i = 0; i < GLIDER_PERIOD; ++i) {
	const cell_t *p = up ? ptr : decr_ptr(ptr);
	cell_t origin = *p - glider->adjust[i][0];
	coord_t y = GET_Y(*p);

	for (j = 0; j < GLIDER_LENGTH; ++j) {
	    if (*p != origin + glider->adjust[i][j]) { break; }
	    p = up ? incr_ptr(p) : decr_ptr(p);
	}

	if (j == GLIDER_LENGTH) {
	    if ((  up && y + GLIDER_MARGIN <= GET_Y(*p))
	     || (! up && y - GLIDER_MARGIN >= GET_Y(*p))) {
		if (origin_ptr != NULL) {
		    *origin_ptr = origin;
		}
		return i;
	    }
	    break;
	}
    }

    return -1;
}

static Boolean check_glider(
    int count, const cell_t *ptr1, const cell_t *ptr2,
    Boolean up, const struct glider *glider)
{
    int i, j;
    cell_t origin;
    const cell_t *p;

    i = is_glider(ptr1, up, glider, &origin);
    if (i >= 0) {
	for (j = 0; j < count; ++j) {
	    ++i;
	    if (i == GLIDER_PERIOD) {
		origin += glider->next;
		i = 0;
	    }
	}

	p = up ? ptr2 : decr_ptr(ptr2);
	for (j = 0; j < GLIDER_LENGTH; ++j) {
	    if (*p != origin + glider->adjust[i][j]) { break; }
	    p = up ? incr_ptr(p) : decr_ptr(p);
	}

	if (j == GLIDER_LENGTH) {
	    return true;
	}
    }

    return false;
}

static int is_twin_glider(
    const cell_t *ptr, Boolean up, const struct glider *glider1,
    const struct glider *glider2, cell_t *origin1_ptr, cell_t *origin2_ptr)
{
    int i, j1, j2;

    for (i = 0; i < GLIDER_PERIOD; ++i) {
	const cell_t *p = up ? ptr : decr_ptr(ptr);
	cell_t origin1 = *p - glider1->adjust[i][0];
	cell_t origin2 = SENTINEL;
	coord_t y = GET_Y(*p);

	for (j1 = j2 = 0; j2 < GLIDER_LENGTH; ) {
	    if (j1 < GLIDER_LENGTH
	     && *p == origin1 + glider1->adjust[i][j1]) {
		++j1;
	    } else if (j2 == 0 && GET_Y(*p) == y) {
		origin2 = *p - glider2->adjust[i][0];
		++j2;
	    } else if (j2 > 0 && j2 < j1
	     && *p == origin2 + glider2->adjust[i][j2]) {
		++j2;
	    } else {
		break;
	    }
	    p = up ? incr_ptr(p) : decr_ptr(p);
	}

	if (j2 == GLIDER_LENGTH) {
	    if ((  up && y + GLIDER_MARGIN <= GET_Y(*p))
	     || (! up && y - GLIDER_MARGIN >= GET_Y(*p))) {
		if (origin1_ptr != NULL && origin2_ptr != NULL) {
		    *origin1_ptr = origin1;
		    *origin2_ptr = origin2;
		}
		return i;
	    }
	    break;
	}
    }

    return -1;
}

static Boolean check_twin_glider(
    int count, const cell_t *ptr1, const cell_t *ptr2,
    Boolean up, const struct glider *glider1, const struct glider *glider2)
{
    int i, j, j1, j2;
    cell_t origin1, origin2;
    const cell_t *p;

    i = is_twin_glider(ptr1, up, glider1, glider2, &origin1, &origin2);
    if (i >= 0) {
	for (j = 0; j < count; ++j) {
	    ++i;
	    if (i == GLIDER_PERIOD) {
		origin1 += glider1->next;
		origin2 += glider2->next;
		i = 0;
	    }
	}

	p = up ? ptr2 : decr_ptr(ptr2);
	for (j1 = j2 = 0; j2 < GLIDER_LENGTH; ) {
	    if (j1 < GLIDER_LENGTH
	     && *p == origin1 + glider1->adjust[i][j1]) {
		++j1;
	    } else if (j2 < j1
	     && *p == origin2 + glider2->adjust[i][j2]) {
		++j2;
	    } else {
		break;
	    }
	    p = up ? incr_ptr(p) : decr_ptr(p);
	}

	if (j2 == GLIDER_LENGTH) {
	    return true;
	}
    }

    return false;
}

static int compare(
    int count, const cell_t *begin, const cell_t *end,
    const cell_t *life_begin, const cell_t *life_end)
{
    offset_t length;
    int i, num, num_gliders = 0;
    Boolean up;
    const cell_t *p, *q;

    while ((length = count_length(begin, end)) > GLIDER_LENGTH) {
	if (check_glider(count, begin, life_begin, true, &glider_up_left)
	 || check_glider(count, begin, life_begin, true, &glider_up_right)) {
	    num = 1;
	    up = true;
	    goto found;
	}

	if (length > GLIDER_LENGTH * 2
	 && check_twin_glider(count, begin, life_begin,
		true, &glider_up_left, &glider_up_right)) {
	    num = 2;
	    up = true;
	    goto found;
	}

	if (check_glider(count, end, life_end, false, &glider_down_left)
	 || check_glider(count, end, life_end, false, &glider_down_right)) {
	    num = 1;
	    up = false;
	    goto found;
	}

	if (length > GLIDER_LENGTH * 2
	 && check_twin_glider(count, end, life_end,
		false, &glider_down_right, &glider_down_left)) {
	    num = 2;
	    up = false;
	    goto found;
	}

	break;

    found:
	for (i = 0; i < GLIDER_LENGTH * num; ++i) {
	    if (up) {
		begin = incr_ptr(begin);
		life_begin = incr_ptr(life_begin);
	    } else {
		end = decr_ptr(end);
		life_end = decr_ptr(life_end);
	    }
	}
	num_gliders += num;
    }

    p = begin;
    q = life_begin;
    for ( ; p != end; p = incr_ptr(p), q = incr_ptr(q)) {
	if (*p != *q) { break; }
    }
    if (p == end && q == life_end) {
	return num_gliders;
    }

    return -1;
}

int check(const cell_t *life_begin, const cell_t *life_end, int *num_gliders)
{
    int count = 0;
    const cell_t *begin, *end;
    offset_t life_length = count_length(life_begin, life_end);

    begin = life_begin;
    end = life_end;
    while ((begin = (const cell_t *) *(end = decr_ptr(begin))) != NULL) {
	++count;
	if (count_length(begin, end) == life_length
	 && (*num_gliders = compare(
		count, begin, end, life_begin, life_end)) >= 0) {
	    return count;
	}
    }

    return 0;
}

void get_bounds(const cell_t *life_begin, const cell_t *life_end,
    coord_t *min_x, coord_t *min_y, coord_t *max_x, coord_t *max_y)
{
    offset_t length;
    int i, num;
    Boolean up;
    const cell_t *p;
    coord_t x, y;

    while ((length = count_length(life_begin, life_end)) > GLIDER_LENGTH) {
	if (is_glider(life_begin, true, &glider_up_left, NULL) >= 0
	 || is_glider(life_begin, true, &glider_up_right, NULL) >= 0) {
	    num = 1;
	    up = true;
	    goto found;
	}

	if (length > GLIDER_LENGTH * 2
	 && is_twin_glider(life_begin, true,
		&glider_up_left, &glider_up_right, NULL, NULL) >= 0) {
	    num = 2;
	    up = true;
	    goto found;
	}

	if (is_glider(life_end, false, &glider_down_left, NULL) >= 0
	 || is_glider(life_end, false, &glider_down_right, NULL) >= 0) {
	    num = 1;
	    up = false;
	    goto found;
	}

	if (length > GLIDER_LENGTH * 2
	 && is_twin_glider(life_end, false,
		&glider_down_right, &glider_down_left, NULL, NULL) >= 0) {
	    num = 2;
	    up = false;
	    goto found;
	}

	break;

    found:
	for (i = 0; i < GLIDER_LENGTH * num; ++i) {
	    if (up) {
		life_begin = incr_ptr(life_begin);
	    } else {
		life_end = decr_ptr(life_end);
	    }
	}
    }

    if (life_begin == life_end) {
	*min_x = *max_x = 0;
	*min_y = *max_y = 0;
	return;
    }

    *min_x = *max_x = GET_X(*life_begin);
    *min_y = *max_y = GET_Y(*life_begin);

    for (p = life_begin; p != life_end; p = incr_ptr(p)) {
	x = GET_X(*p);
	if (x < *min_x) { *min_x = x; }
	if (x > *max_x) { *max_x = x; }

	y = GET_Y(*p);
	if (y < *min_y) { *min_y = y; }
	if (y > *max_y) { *max_y = y; }
    }
}
