/* life.c */

/*
 * Copyright (c) 1997-2002 Hoshi Takanori
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <PalmOS.h>

#ifdef USE_SONY_HRLIB
#include <SonySystemFtr.h>
#include <SonyHRLib.h>
#endif

#ifdef USE_HANDERA_VGA
#include <Vga.h>
#endif

#include "life.h"
#include "life_rcp.h"


/*
 * macros
 */

#define MAX_LENGTH(max) (((max) - ((max) / 4 + 6144)) / sizeof(cell_t))
#define LENGTH_DELTA 256

#define SCROLL_MARGIN 20
#define SCROLL_OFFSET 16

#define MIN_MAGNIFY -6
#define MAX_MAGNIFY 4

#define scroll_x prefs.scroll_x
#define scroll_y prefs.scroll_y
#define magnify prefs.magnify

#define gadget_x gadget_bounds.topLeft.x
#define gadget_y gadget_bounds.topLeft.y
#define gadget_width gadget_bounds.extent.x
#define gadget_height gadget_bounds.extent.y


/*
 * global variables
 */

UInt32 rom_version;

struct prefs prefs;

static cell_t *tape = NULL;
static cell_t *tape_begin;
static cell_t *tape_end;
static offset_t tape_length;
static offset_t init_length;
static const cell_t *life_begin;
static const cell_t *life_end;

static gen_t generation;
static gen_t until;
static Boolean time;
static gen_t from_gen;
static UInt32 from_ticks;

static Boolean running;
static Boolean can_back;

static RectangleType gadget_bounds;

static char gens[12];
static char cells[12];
static char mags[6];

static RectangleType cell_rect;
static coord_t cell_mag;
static coord_t delta_x;
static coord_t delta_y;
static coord_t prev_y;

static Boolean dragging;
static coord_t start_x;
static coord_t start_y;
static coord_t old_x;
static coord_t old_y;

#ifdef USE_SONY_HRLIB
static UInt16 sony_hrlib = sysInvalidRefNum;
#endif

#ifdef USE_HANDERA_VGA
Boolean handera_vga;
static coord_t screen_width = 240;
static coord_t screen_height = 240;
#endif


/*
 * init, free and reset life cells
 */

static void init_life(void)
{
    UInt32 free, max;
    offset_t cur_length;
#ifdef USE_SONY_HRLIB
    SonySysFtrSysInfoP info;
    UInt32 width, height;
#endif
#ifdef USE_HANDERA_VGA
    UInt32 vga_version;
#endif

    FtrGet(sysFtrCreator, sysFtrNumROMVersion, &rom_version);

    if (PrefGetAppPreferencesV10(
	    CREATOR, PREFS_VERSION, &prefs, sizeof prefs) == false) {
	if (PrefGetAppPreferencesV10(CREATOR, OLD_PREFS_VERSION,
		&prefs, sizeof(struct old_prefs)) == false) {
	    MemSet(&prefs, sizeof prefs, 0);

	    prefs.rand_width = 32;
	    prefs.rand_height = 32;
	    prefs.rand_density = 24;
	    prefs.rand_radius = 24;
	    prefs.rand_density1 = 32;
	    prefs.rand_density2 = 16;
	}

	prefs.vfs_top = 0;
	prefs.vfs_selection = 0;
	MemSet(prefs.vfs_name, VFS_NAME_LENGTH, 0);
    }

    if (magnify < MIN_MAGNIFY) {
	magnify = MIN_MAGNIFY;
    } else if (magnify > MAX_MAGNIFY) {
	magnify = MAX_MAGNIFY;
    }

    prefs.vfs_name[VFS_NAME_LENGTH - 1] = '\0';

#ifdef USE_SONY_HRLIB
    if (FtrGet(sonySysFtrCreator, sonySysFtrNumSysInfoP, (UInt32 *) &info)
	    == errNone
     && info != NULL
     && (info->libr & sonySysFtrSysInfoLibrHR) != 0) {
	if (SysLibFind(sonySysLibNameHR, &sony_hrlib) != errNone
	 && SysLibLoad(sysFileTLibrary, sonySysFileCHRLib, &sony_hrlib)
		!= errNone) {
	    sony_hrlib = sysInvalidRefNum;
	}
    }

    if (sony_hrlib != sysInvalidRefNum) {
	if (HROpen(sony_hrlib) == errNone) {
	    width = hrWidth;
	    height = hrHeight;
	    if (HRWinScreenMode(sony_hrlib, winScreenModeSet,
		    &width, &height, NULL, NULL) != errNone) {
		HRClose(sony_hrlib);
		sony_hrlib = sysInvalidRefNum;
	    }
	} else {
	    sony_hrlib = sysInvalidRefNum;
	}
    }
#endif

#ifdef USE_HANDERA_VGA
    if (FtrGet(TRGSysFtrID, TRGVgaFtrNum, &vga_version) == errNone
     && sysGetROMVerMajor(vga_version) >= 1) {
	VgaSetScreenMode(screenMode1To1, rotateModeNone);
	handera_vga = true;
    }
#endif

    MemHeapFreeBytes(0, &free, &max);
    tape_length = MAX_LENGTH(max) * 2;

    while (tape == NULL) {
	if (tape_length > LENGTH_DELTA * 2) {
	    tape_length -= LENGTH_DELTA;
	} else {
	    tape_length /= 2;
	}
	tape = MemPtrNew(sizeof(cell_t) * tape_length);
    }

    read_db(tape, tape_length, &init_length, &generation, &cur_length);

    tape[init_length] = NULL;
    tape_begin = tape + init_length;
    tape_end = tape + tape_length;
    life_begin = tape_begin + 1;
    if (generation > 0) {
	life_end = life_begin + cur_length;
    } else {
	life_end = life_begin + init_length;
	MemMove((cell_t *) life_begin, tape, sizeof(cell_t) * init_length);
    }

    running = false;
}

static void free_life(void)
{
    PrefSetAppPreferencesV10(CREATOR, PREFS_VERSION, &prefs, sizeof prefs);

    write_db(tape, init_length, generation, life_begin, life_end);

    if (tape != NULL) {
	MemPtrFree(tape);
    }

#ifdef USE_SONY_HRLIB
    if (sony_hrlib != sysInvalidRefNum) {
	HRWinScreenMode(sony_hrlib,
		winScreenModeSetToDefaults, NULL, NULL, NULL, NULL);
	HRClose(sony_hrlib);
    }
#endif
}

void reset_life(offset_t length)
{
    init_length = length;

    tape[init_length] = NULL;
    tape_begin = tape + init_length;
    life_begin = tape_begin + 1;
    life_end = life_begin + init_length;
    MemMove((cell_t *) life_begin, tape, sizeof(cell_t) * init_length);
    generation = 0;

    scroll_x = 0;
    scroll_y = 0;
    running = false;
}

offset_t count_length(const cell_t *begin, const cell_t *end)
{
    if (begin <= end) {
	return end - begin;
    } else {
	return (end - tape_begin) + (tape_end - begin);
    }
}

const cell_t *incr_ptr(const cell_t *ptr)
{
    return (ptr + 1 == tape_end) ? tape_begin : ptr + 1;
}

const cell_t *decr_ptr(const cell_t *ptr)
{
    return (ptr == tape_begin) ? tape_end - 1 : ptr - 1;
}


/*
 * drawing, calculating and scrolling life cells
 */

static void erase_prev(coord_t y)
{
    RectangleType r;

    if (y > prev_y) {
#ifdef USE_SONY_HRLIB
	if (sony_hrlib == sysInvalidRefNum) {
#endif
	RctSetRectangle(&r, gadget_x, prev_y + 1, gadget_width, y - prev_y);
	WinEraseRectangle(&r, 0);
#ifdef USE_SONY_HRLIB
	} else {
	    RctSetRectangle(&r,
		    gadget_x * 2, prev_y + 1, gadget_width * 2, y - prev_y);
	    HRWinEraseRectangle(sony_hrlib, &r, 0);
	}
#endif
	prev_y = y;
    }
}

static void invert_cell(cell_t cell)
{
    coord_t x = GET_X(cell) + scroll_x;
    coord_t y = GET_Y(cell) + scroll_y;

    if (magnify == 0) {
	x += gadget_x + gadget_width / 2;
	y += gadget_y + gadget_height / 2;
	if (x >= gadget_x && x < gadget_x + gadget_width
	 && y >= gadget_y && y < gadget_y + gadget_height) {
	    if (rom_version >= sysMakeROMVersion(3, 5, 0, 0, 0)) {
		WinInvertPixel(x, y);
	    } else {
		WinInvertLine(x, y, x, y);
	    }
	}
    } else if (magnify > 0) {
	x = (x << magnify) - delta_x + gadget_x + gadget_width / 2;
	y = (y << magnify) - delta_y + gadget_y + gadget_height / 2;
	if (x >= gadget_x && x + cell_mag <= gadget_x + gadget_width
	 && y >= gadget_y && y + cell_mag <= gadget_y + gadget_height) {
	    cell_rect.topLeft.x = x;
	    cell_rect.topLeft.y = y;
	    WinInvertRectangle(&cell_rect, 0);
	}
#ifdef USE_SONY_HRLIB
    } else if (sony_hrlib == sysInvalidRefNum) {
#else
    } else {
#endif
	x = (x >> (- magnify)) + gadget_x + gadget_width / 2;
	y = (y >> (- magnify)) + gadget_y + gadget_height / 2;
	if (x >= gadget_x && x < gadget_x + gadget_width
	 && y >= gadget_y && y < gadget_y + gadget_height) {
	    if (y > prev_y) {
		erase_prev(y);
	    }
	    if (rom_version >= sysMakeROMVersion(3, 5, 0, 0, 0)) {
		WinDrawPixel(x, y);
	    } else {
		WinDrawLine(x, y, x, y);
	    }
	}
#ifdef USE_SONY_HRLIB
    } else if (magnify == -1) {
	x += gadget_x * 2 + gadget_width;
	y += gadget_y * 2 + gadget_height;
	if (x >= gadget_x * 2 && x < (gadget_x + gadget_width) * 2
	 && y >= gadget_y * 2 && y < (gadget_y + gadget_height) * 2) {
	    HRWinInvertPixel(sony_hrlib, x, y);
	}
    } else {
	x = (x >> (- magnify - 1)) + gadget_x * 2 + gadget_width;
	y = (y >> (- magnify - 1)) + gadget_y * 2 + gadget_height;
	if (x >= gadget_x * 2 && x < (gadget_x + gadget_width) * 2
	 && y >= gadget_y * 2 && y < (gadget_y + gadget_height) * 2) {
	    if (y > prev_y) {
		erase_prev(y);
	    }
	    HRWinDrawPixel(sony_hrlib, x, y);
	}
#endif
    }
}

static void draw_cells(Boolean update)
{
    const cell_t *ptr, *prev_begin, *prev_end, *prev_ptr;
    cell_t saved_end, cell, prev_cell, sentinel;

    if (! update) {
	WinDrawRectangleFrame(rectangleFrame, &gadget_bounds);
	WinEraseRectangle(&gadget_bounds, 0);
    }

    if (magnify > 0) {
	cell_rect.extent.x = cell_rect.extent.y = cell_mag = 1 << magnify;
	delta_x = ((gadget_width  >> magnify) & 1) ? cell_mag / 2 : 0;
	delta_y = ((gadget_height >> magnify) & 1) ? cell_mag / 2 : 0;
    }

    if (! update || magnify < 0) {
	prev_begin = prev_end = &sentinel;
#ifdef USE_SONY_HRLIB
	if (sony_hrlib == sysInvalidRefNum) {
#endif
	prev_y = gadget_y - 1;
#ifdef USE_SONY_HRLIB
	} else if (! update || magnify < -1) {
	    prev_y = gadget_y * 2 - 1;
	} else {
	    prev_end = decr_ptr(life_begin);
	    prev_begin = (const cell_t *) *prev_end;
	}
#endif
    } else {
	prev_end = decr_ptr(life_begin);
	prev_begin = (const cell_t *) *prev_end;
    }

    saved_end = *life_end;
    *(cell_t *) life_end = SENTINEL;
    *(cell_t *) prev_end = SENTINEL;

    cell = *(ptr = life_begin);
    prev_cell = *(prev_ptr = prev_begin);

    while (true) {
	if (cell < prev_cell) {
	    invert_cell(cell);
	    cell = *(ptr = incr_ptr(ptr));
	} else if (cell > prev_cell) {
	    invert_cell(prev_cell);
	    prev_cell = *(prev_ptr = incr_ptr(prev_ptr));
	} else if (cell != SENTINEL) {
	    cell = *(ptr = incr_ptr(ptr));
	    prev_cell = *(prev_ptr = incr_ptr(prev_ptr));
	} else {
	    break;
	}
    }

    if (magnify < 0) {
#ifdef USE_SONY_HRLIB
	if (sony_hrlib == sysInvalidRefNum) {
#endif
	erase_prev(gadget_y + gadget_height - 1);
#ifdef USE_SONY_HRLIB
	} else if (magnify < -1) {
	    erase_prev((gadget_y + gadget_height) * 2 - 1);
	}
#endif
    }

    *(cell_t *) life_end = saved_end;
    *(cell_t *) prev_end = (cell_t) prev_begin;

    SetFieldNumber((FieldPtr) GetObjectPtr(main_gens_id), gens, generation);
    SetFieldNumber((FieldPtr) GetObjectPtr(main_cells_id),
	cells, count_length(life_begin, life_end));

    if (magnify >= 0) {
	mags[0] = 'x';
	StrIToA(&mags[1], 1 << magnify);
    } else {
	mags[0] = '/';
	StrIToA(&mags[1], 1 << (- magnify));
    }
    FldSetTextPtr((FieldPtr) GetObjectPtr(main_mags_id), mags);
    FldDrawField((FieldPtr) GetObjectPtr(main_mags_id));
}

static void draw_bitmap(int bitmap_id, int button_id, coord_t offset)
{
    MemHandle h;
    BitmapPtr p;
    RectangleType bounds;

#ifdef USE_HANDERA_VGA
    if (handera_vga) {
	h = DmGetResource(bitmapRsc,
		bitmap_id + (arrows_vga_bitmap_id - arrows_bitmap_id));
    } else {
#endif
    h = DmGetResource(bitmapRsc, bitmap_id);
#ifdef USE_HANDERA_VGA
    }
#endif
    if (h != NULL) {
	p = MemHandleLock(h);
	GetObjectBounds(button_id, &bounds);
	WinDrawBitmap(p, bounds.topLeft.x + offset, bounds.topLeft.y + offset);
	MemHandleUnlock(h);
	DmReleaseResource(h);
    }
}

static void draw_go_stop_button(void)
{
    draw_bitmap(running ? stop_bitmap_id : go_bitmap_id, main_go_id, 0);
}

static void draw_back_button(Boolean forced)
{
    ControlPtr back_btn;
    Boolean new_back;

    new_back = (*decr_ptr(life_begin) != NULL);
    if (forced || new_back != can_back) {
	can_back = new_back;

	back_btn = (ControlPtr) GetObjectPtr(main_back_id);
	if (! forced && ! can_back) {
	    CtlEraseControl(back_btn);
	}
	CtlSetUsable(back_btn, can_back);
	if (! forced && can_back) {
	    CtlDrawControl(back_btn);
	}

	draw_bitmap(can_back ? back_bitmap_id : cant_back_bitmap_id,
		main_back_id, 0);
    }
}

static void set_running(Boolean run)
{
    ControlPtr go_btn, stop_btn;

    if (run != running) {
	go_btn = (ControlPtr) GetObjectPtr(main_go_id);
	stop_btn = (ControlPtr) GetObjectPtr(main_stop_id);

	CtlEraseControl(run ? go_btn : stop_btn);
	CtlSetUsable(go_btn, ! run);
	CtlSetUsable(stop_btn, run);
	CtlDrawControl(run ? stop_btn : go_btn);
	running = run;

	draw_go_stop_button();
    }
}

static void go_until(void)
{
    FormPtr dialog;
    FieldPtr field;
    int button;

    dialog = FrmInitForm(until_form_id);
#ifdef USE_HANDERA_VGA
    if (handera_vga) {
	VgaFormModify(dialog, vgaFormModify160To240);
    }
#endif
    FrmSetFocus(dialog, FrmGetObjectIndex(dialog, until_field_id));
    SetFieldNumHandle(dialog, until_field_id, generation);
    field = GetObjectPtrIn(dialog, until_field_id);
    FldSetSelection(field, 0, FldGetTextLength(field));

    button = FrmDoDialog(dialog);
    if (button == until_ok_id) {
	until = GetFieldNumber(dialog, until_field_id);
	time = CtlGetValue(GetObjectPtrIn(dialog, until_time_id));
	from_gen = generation;
	from_ticks = TimGetTicks();
    }
    FrmDeleteForm(dialog);

    if (button == until_ok_id) {
	set_running(until > generation);
    }
}

static void show_time(void)
{
    char from[6], to[6], ticks[12];

    StrIToA(from, from_gen);
    StrIToA(to, generation);
    StrIToA(ticks, TimGetTicks() - from_ticks);
    FrmCustomAlert(time_alert_id, from, to, ticks);
}

static void next_gen(Boolean need_check)
{
    const cell_t *begin, *end;
    cell_t *ptr;

    if (life_end == life_begin) {
	set_running(false);
	draw_back_button(false);
	return;
    }

    begin = incr_ptr(life_end);
    ptr = (cell_t *) decr_ptr(life_begin);
    end = merge(tape_begin, tape_end, life_begin, life_end,
		(cell_t *) begin, ptr);
    if (end == NULL) {
	*ptr = NULL;
	set_running(false);
	draw_back_button(false);
	FrmAlert(outof_error_id);
	return;
    }

    *(cell_t *) life_end = (cell_t) life_begin;

    life_begin = begin;
    life_end = end;

    for ( ; *ptr != NULL; ptr = (cell_t *) begin) {
	begin = decr_ptr((const cell_t *) *ptr);
	if ((begin < ptr && (end > begin && end <= ptr))
	 || (begin > ptr && (end <= ptr || end > begin))) {
	    break;
	}
    }
    *ptr = NULL;

    if (need_check) {
	int count, num_gliders;
	char gens[6], from[6], gliders[6];

	count = check(life_begin, life_end, &num_gliders);
	if (count > 0) {
	    life_end = decr_ptr(life_begin);
	    life_begin = (const cell_t *) *life_end;
	    set_running(false);

	    StrIToA(gens, count);
	    StrIToA(from, generation + 1 - count);
	    if (num_gliders == 0) {
		FrmCustomAlert(loop_alert_id, gens, from, " ");
	    } else {
		StrIToA(gliders, num_gliders);
		FrmCustomAlert(loop2_alert_id, gens, from, gliders);
	    }

	    return;
	}
    }

    ++generation;

    draw_back_button(false);
    draw_cells(true);
}

static void go_back(void)
{
    const cell_t *ptr = decr_ptr(life_begin);

    if (*ptr != NULL) {
	--generation;

	draw_cells(true);

	life_begin = (const cell_t *) *ptr;
	life_end = ptr;

	SetFieldNumber((FieldPtr) GetObjectPtr(main_cells_id),
	    cells, count_length(life_begin, life_end));

	draw_back_button(false);
    }
}

static Boolean scroll(EventPtr event)
{
    coord_t x = event->screenX;
    coord_t y = event->screenY;
    coord_t offset;
    Boolean handled = false;

    if (magnify >= 0) {
	offset = SCROLL_OFFSET >> magnify;
    } else {
	offset = SCROLL_OFFSET << (- magnify);
    }

    if (RctPtInRectangle(x, y, &gadget_bounds)) {
	if (x < gadget_x + SCROLL_MARGIN) {
	    scroll_x += offset;
	    handled = true;
	} else if (x >= gadget_x + gadget_width - SCROLL_MARGIN) {
	    scroll_x -= offset;
	    handled = true;
	}

	if (y < gadget_y + SCROLL_MARGIN) {
	    scroll_y += offset;
	    handled = true;
	} else if (y >= gadget_y + gadget_height - SCROLL_MARGIN) {
	    scroll_y -= offset;
	    handled = true;
	}

	if (handled) {
	    draw_cells(false);
	} else {
	    dragging = true;
	    start_x = old_x = x;
	    start_y = old_y = y;
	    handled = true;
	}
    }

    return handled;
}

static void scroll_rect(EventPtr event)
{
    coord_t new_x = event->screenX;
    coord_t new_y = event->screenY;
    RectangleType erase;

    if (new_x < old_x) {
	WinScrollRectangle(&gadget_bounds, winLeft, old_x - new_x, &erase);
	WinEraseRectangle(&erase, 0);
    } else if (new_x > old_x) {
	WinScrollRectangle(&gadget_bounds, winRight, new_x - old_x, &erase);
	WinEraseRectangle(&erase, 0);
    }

    if (new_y < old_y) {
	WinScrollRectangle(&gadget_bounds, winUp, old_y - new_y, &erase);
	WinEraseRectangle(&erase, 0);
    } else if (new_y > old_y) {
	WinScrollRectangle(&gadget_bounds, winDown, new_y - old_y, &erase);
	WinEraseRectangle(&erase, 0);
    }

    old_x = new_x;
    old_y = new_y;
}

static void show_whole(void)
{
    coord_t min_x, min_y, max_x, max_y, width, height;

    get_bounds(life_begin, life_end, &min_x, &min_y, &max_x, &max_y);
    scroll_x = - (min_x + max_x) >> 1;
    scroll_y = - (min_y + max_y) >> 1;

    for (magnify = MAX_MAGNIFY; magnify > MIN_MAGNIFY; --magnify) {
	width = max_x - min_x + 1;
	height = max_y - min_y + 1;
	if (magnify > 0) {
	    width <<= magnify;
	    height <<= magnify;
	} else if (magnify < 0) {
	    width >>= (- magnify);
	    height >>= (- magnify);
	}
	if (width < gadget_width && height < gadget_height) {
	    break;
	}
    }
}


/*
 * event handlers
 */

#ifdef USE_HANDERA_VGA

static void adjust_object(
	FormPtr form, int id, Boolean move, coord_t delta_x, coord_t delta_y)
{
    int index;
    RectangleType r;

    index = FrmGetObjectIndex(form, id);
    FrmGetObjectBounds(form, index, &r);
    if (move) {
	r.topLeft.x += delta_x;
	r.topLeft.y += delta_y;
    } else {
	r.extent.x += delta_x;
	r.extent.y += delta_y;
    }
    FrmSetObjectBounds(form, index, &r);
}

static void resize_main_form(FormPtr form)
{
    RectangleType r;
    coord_t delta_x, delta_y;
    int id;

    WinGetDisplayExtent(&r.extent.x, &r.extent.y);
    delta_x = r.extent.x - screen_width;
    delta_y = r.extent.y - screen_height;

    if (delta_x != 0 || delta_y != 0) {
	r.topLeft.x = 0;
	r.topLeft.y = 0;

	if (delta_x == 0 && FrmVisible(form)) {
	    WinEraseRectangle(&r, 0);
	}

	WinSetWindowBounds(FrmGetWindowHandle(form), &r);

	adjust_object(form, main_gadget_id, false, delta_x, delta_y);
	for (id = main_gens_id; id <= main_mags_id; ++id) {
	    adjust_object(form, id, true, delta_x / 2, delta_y);
	}

	screen_width = r.extent.x;
	screen_height = r.extent.y;
    }
}

#endif

static Boolean handle_main_form_event(EventPtr event)
{
    Boolean handled = false;
#ifdef USE_HANDERA_VGA
    FormPtr form;
    int id;
#endif

    switch (event->eType) {
    case frmOpenEvent:
#ifdef USE_HANDERA_VGA
	if (handera_vga) {
	    form = FrmGetActiveForm();
	    VgaFormModify(form, vgaFormModify160To240);
	    adjust_object(form, main_gadget_id, false, 0, -1);
	    for (id = main_reset_id; id <= main_whole_id; ++id) {
		adjust_object(form, id, true, 0, -1);
	    }
	}

	/* fall through */

    case displayExtentChangedEvent:
	if (handera_vga) {
	    resize_main_form(FrmGetActiveForm());
	}
#endif
	GetObjectBounds(main_gadget_id, &gadget_bounds);
	FrmDrawForm(FrmGetActiveForm());
	dragging = false;
	handled = true;

	/* fall through */

    case frmUpdateEvent:
	draw_bitmap(arrows_bitmap_id, main_reset_id, -1);
	draw_bitmap(magnify_bitmap_id, main_minus_id, -1);
	draw_go_stop_button();
	draw_back_button(true);
	draw_cells(false);
	break;

    case penDownEvent:
	handled = scroll(event);
	break;

    case penMoveEvent:
	if (dragging) {
	    scroll_rect(event);
	    handled = true;
	}
	break;

    case penUpEvent:
	if (dragging) {
	    dragging = false;
	    if (magnify >= 0) {
		coord_t mag = 1 << magnify;
		scroll_x += (event->screenX - start_x + mag / 2) >> magnify;
		scroll_y += (event->screenY - start_y + mag / 2) >> magnify;
	    } else {
		scroll_x += (event->screenX - start_x) << (- magnify);
		scroll_y += (event->screenY - start_y) << (- magnify);
	    }
	    draw_cells(false);
	    handled = true;
	}
	break;

    case menuEvent:
	MenuEraseStatus(NULL);

	switch (event->data.menu.itemID) {
	case main_load_item:
	    set_running(false);
	    load_memo(tape, tape_length);
	    draw_back_button(false);
	    draw_cells(false);
	    handled = true;
	    break;

	case main_load_vfs_item:
	    set_running(false);
	    load_vfs(tape, tape_length);
	    draw_back_button(false);
	    draw_cells(false);
	    handled = true;
	    break;

	case main_random_item:
	    set_running(false);
	    random_life(tape, tape_length);
	    draw_back_button(false);
	    draw_cells(false);
	    handled = true;
	    break;

	case main_about_item:
	    FrmAlert(about_alert_id);
	    handled = true;
	    break;
	}
	break;

    case ctlSelectEvent:
	switch (event->data.ctlSelect.controlID) {
	case main_go_id:
	    until = -1;
	    set_running(true);
	    handled = true;
	    break;

	case main_stop_id:
	    set_running(false);
	    handled = true;
	    break;

	case main_until_id:
	    go_until();
	    handled = true;
	    break;

	case main_minus_id:
	    if (magnify > MIN_MAGNIFY) {
		--magnify;
		draw_cells(false);
	    }
	    handled = true;
	    break;

	case main_plus_id:
	    if (magnify < MAX_MAGNIFY) {
		++magnify;
		draw_cells(false);
	    }
	    handled = true;
	    break;

	case main_whole_id:
	    if (life_begin != life_end) {
		show_whole();
		draw_cells(false);
	    }
	    handled = true;
	    break;

	case main_reset_id:
	    set_running(false);
	    reset_life(init_length);
	    draw_back_button(false);
	    draw_cells(false);
	    handled = true;
	    break;
	}
	break;

    case ctlRepeatEvent:
	switch (event->data.ctlRepeat.controlID) {
	case main_step_id:
	    set_running(false);
	    next_gen(false);
	    break;

	case main_back_id:
	    set_running(false);
	    go_back();
	    break;
	}
	break;

    case nilEvent:
	if (running && ! dragging && ! IsMenuVisible()) {
	    next_gen(true);
	    if (until >= 0 && generation >= until) {
		set_running(false);
		if (time) {
		    show_time();
		}
	    }
	}
	handled = true;
	break;

    default:
	break;
    }

    return handled;
}

static Boolean handle_app_event(EventPtr event)
{
    FormPtr form;

    if (event->eType == frmLoadEvent) {
	form = FrmInitForm(event->data.frmLoad.formID);
	FrmSetActiveForm(form);
	if (event->data.frmLoad.formID == main_form_id) {
	    FrmSetEventHandler(form, handle_main_form_event);
	}
	return true;
    }
    return false;
}

static void event_loop(void)
{
    EventType event;
    Err error;

    do {
	EvtGetEvent(&event,
	    (running && ! dragging && ! IsMenuVisible()) ? 1 : evtWaitForever);
	if (SysHandleEvent(&event) == false
	 && MenuHandleEvent(NULL, &event, &error) == false
	 && handle_app_event(&event) == false) {
	    FrmDispatchEvent(&event);
	}
    } while (event.eType != appStopEvent);
}


/*
 * main entry point
 */

UInt32 PilotMain(UInt16 cmd, void *cmdPBP, UInt16 launchFlags)
{
    if (cmd == sysAppLaunchCmdNormalLaunch) {
	init_life();
	FrmGotoForm(main_form_id);
	event_loop();
	FrmSaveAllForms();
	FrmCloseAllForms();
	free_life();
    }
    return 0;
}
