/* load.c */

/*
 * Copyright (c) 1997-2002 Hoshi Takanori
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <PalmOS.h>
#include <VFSMgr.h>

#ifdef USE_HANDERA_VGA
#include <Vga.h>
#endif

#include "life.h"
#include "life_rcp.h"


/*
 * macros
 */

#define is_digit(c) ((c) >= '0' && (c) <= '9')
#define is_newline(c) ((c) == '\n' || (c) == '\r')

#define LIFE_CATEGORY "Life"

#define MAX_FILENAME 256
#define LIFE_DIRECTORY "/PALM/Programs/Life"
#define LIFE_DIRECTORY_LENGTH 19


/*
 * global variables
 */

static DmOpenRef memo_db;
static UInt16 life_category;
static ListPtr list;
static Boolean last_inited;


/*
 * load functions
 */

static Boolean get_num(const char *str, long strlen, long *pos, coord_t *value)
{
    Boolean neg = false;

    if (*pos + 1 >= strlen || str[*pos] != ' ') {
	return false;
    }

    if (str[++*pos] == '-') {
	neg = true;
	if (++*pos >= strlen) {
	    return false;
	}
    }

    if (! is_digit(str[*pos])) {
	return false;
    }

    *value = 0;
    do {
	*value = *value * 10 + str[*pos] - '0';
    } while (++*pos < strlen && is_digit(str[*pos]));

    if (neg) {
	*value = - *value;
    }

    return true;
}

static offset_t load_life(const char *str, long strlen, cell_t *tape)
{
    long pos = 0;
    offset_t length = 0;
    coord_t x = 0, y = 0;

    while (pos < strlen) {
	if (str[pos] == '#') {
	    coord_t new_x, new_y;
	    if ((++pos < strlen && (str[pos] == 'P' || str[pos] == 'p'))
	     && ++pos < strlen
	     && get_num(str, strlen, &pos, &new_x)
	     && get_num(str, strlen, &pos, &new_y)) {
		x = new_x;
		y = new_y;
	    }
	} else if (str[pos] == '.' || str[pos] == '*') {
	    coord_t x1 = x;
	    while (pos < strlen && (str[pos] == '.' || str[pos] == '*')) {
		if (str[pos] == '*') {
		    if (tape == NULL) {
			++length;
		    } else {
			cell_t cell = CELL(x1, y);
			offset_t i;
			for (i = length; i > 0 && tape[i - 1] > cell; --i) {}
			if (i == 0 || (i > 0 && tape[i - 1] != cell)) {
			    if (i < length) {
				MemMove(tape + i + 1, tape + i,
					sizeof(cell_t) * (length - i));
			    }
			    tape[i] = cell;
			    ++length;
			}
		    }
		}
		++x1;
		++pos;
	    }
	    ++y;
	}

	while (pos < strlen) {
	    if (is_newline(str[pos])) {
		++pos;
		break;
	    }
	    ++pos;
	}
    }

    return length;
}


/*
 * loading life pattern from memo
 */

static void draw_memo_item(Int16 item_no, RectanglePtr bounds, Char **dummy)
{
    UInt16 index = 0;

    if (DmSeekRecordInCategory(
	    memo_db, &index, item_no, dmSeekForward, life_category) == 0) {
	MemHandle handle = DmQueryRecord(memo_db, index);
	Int16 width = bounds->extent.x;
	Int16 length = MemHandleSize(handle);
	const Char *ptr = MemHandleLock(handle);
	Boolean fits;

	FntCharsInWidth(ptr, &width, &length, &fits);
	WinDrawChars(ptr, length, bounds->topLeft.x, bounds->topLeft.y);

	MemHandleUnlock(handle);
    }
}

static void scroll_list(WinDirectionType direction)
{
    int top, visible;

    if (rom_version >= sysMakeROMVersion(2, 0, 0, 0, 0)) {
	visible = LstGetVisibleItems(list);
    } else {
	visible = list_visible_items;
    }

    if (direction == winUp) {
	top = GetTopItem(list) - (visible - 1);
	if (top < 0) {
	    top = 0;
	}
    } else {
	top = GetTopItem(list) + (visible - 1);
	if (top > LstGetNumberOfItems(list) - visible) {
	    top = LstGetNumberOfItems(list) - visible;
	}
    }

    if (top != GetTopItem(list)) {
	LstSetTopItem(list, top);
	LstEraseList(list);
	LstDrawList(list);
    }
}

static Boolean handle_load_dialog_event(EventPtr event)
{
    Boolean handled = false;

    switch (event->eType) {
    case keyDownEvent:
	if (event->data.keyDown.chr == pageUpChr) {
	    scroll_list(winUp);
	    handled = true;
	} else if (event->data.keyDown.chr == pageDownChr) {
	    scroll_list(winDown);
	    handled = true;
	}
	break;

    default:
	break;
    }

    return handled;
}

void load_memo(cell_t *tape, offset_t tape_length)
{
    FormPtr dialog;
    UInt16 numRecords, i, index;
    MemHandle handle;
    int button, selection;

    memo_db = DmOpenDatabaseByTypeCreator('DATA', sysFileCMemo, dmModeReadOnly);
    if (memo_db == 0) {
	FrmAlert(load_error1_id);
	return;
    }

    life_category = FindCategory(memo_db, LIFE_CATEGORY);
    if (life_category >= dmRecNumCategories) {
	DmCloseDatabase(memo_db);
	FrmAlert(load_error2_id);
	return;
    }

    numRecords = DmNumRecordsInCategory(memo_db, life_category);
    if (numRecords == 0) {
	DmCloseDatabase(memo_db);
	FrmAlert(load_error3_id);
	return;
    }

    dialog = FrmInitForm(load_form_id);
#ifdef USE_HANDERA_VGA
    if (handera_vga) {
	VgaFormModify(dialog, vgaFormModify160To240);
    }
#endif
    list = GetObjectPtrIn(dialog, load_list_id);
    LstSetDrawFunction(list, &draw_memo_item);
    LstSetListChoices(list, NULL, numRecords);
    if (last_inited) {
	LstSetTopItem(list, prefs.last_top);
	LstSetSelection(list, prefs.last_selection);
	LstMakeItemVisible(list, prefs.last_selection);
    } else {
	for (i = 0, index = 0; i < numRecords; ++i, ++index) {
	    handle = DmQueryNextInCategory(memo_db, &index, life_category);
	    if (MemHandleToLocalID(handle) == prefs.last_id) {
		LstSetTopItem(list, prefs.last_top);
		LstSetSelection(list, i);
		LstMakeItemVisible(list, i);
	    }
	}
    }
    FrmSetEventHandler(dialog, handle_load_dialog_event);

    button = FrmDoDialog(dialog);
    selection = LstGetSelection(list);
    prefs.last_top = GetTopItem(list);
    prefs.last_selection = selection;
    index = 0;
    if (selection >= 0 && selection < numRecords
     && DmSeekRecordInCategory(
	    memo_db, &index, selection, dmSeekForward, life_category) == 0) {
	handle = DmQueryRecord(memo_db, index);
	prefs.last_id = MemHandleToLocalID(handle);
    }
    last_inited = true;
    FrmDeleteForm(dialog);

    if (button == load_ok_id && selection != noListSelection) {
	UInt16 index = 0;

	if (DmSeekRecordInCategory(memo_db,
		&index, selection, dmSeekForward, life_category) == 0) {
	    MemHandle handle = DmQueryRecord(memo_db, index);
	    const char *ptr = MemHandleLock(handle);
	    offset_t length;

	    length = load_life(ptr, MemHandleSize(handle), NULL);

	    if (length <= 0) {
		FrmAlert(load_error4_id);
	    } else if (length >= tape_length / 2 - 1) {
		FrmAlert(load_error5_id);
	    } else {
		length = load_life(ptr, MemHandleSize(handle), tape);
		reset_life(length);
	    }

	    MemHandleUnlock(handle);
	}
    }

    DmCloseDatabase(memo_db);
}


/*
 * loading life pattern from vfs
 */

static UInt16 find_volume(void)
{
    UInt32 vfs_version, iterator = vfsIteratorStart;
    UInt16 vol_ref;

    if (FtrGet(sysFileCVFSMgr, vfsFtrIDVersion, &vfs_version) != errNone) {
	return vfsInvalidVolRef;
    }

    if (VFSVolumeEnumerate(&vol_ref, &iterator) == errNone) {
	return vol_ref;
    }

    return vfsInvalidVolRef;
}

static Int16 compare_names(void *s1, void *s2, Int32 other)
{
    return StrCaselessCompare(*(Char **) s1, *(Char **) s2);
}

static int enum_directory(UInt16 vol_ref, const Char *path, Char ***names)
{
    FileRef dir_ref;
    UInt32 iterator;
    FileInfoType info;
    Char name[MAX_FILENAME];
    int max_cnt, cnt, len;
    long max_size, offset;

    *names = NULL;

    if (VFSFileOpen(vol_ref, path, vfsModeRead, &dir_ref) != errNone) {
	return 0;
    }

    max_cnt = 0;
    max_size = 0;
    for (iterator = vfsIteratorStart; iterator != vfsIteratorStop; ) {
	info.nameP = name;
	info.nameBufLen = sizeof name;
	if (VFSDirEntryEnumerate(dir_ref, &iterator, &info) == errNone) {
	    ++max_cnt;
	    max_size += StrLen(name) + 1;
	} else {
	    break;
	}
    }

    VFSFileClose(dir_ref);

    if (max_cnt == 0) {
	return 0;
    }

    *names = MemPtrNew(sizeof(Char *) * max_cnt + max_size);
    if (*names == NULL) {
	return 0;
    }

    if (VFSFileOpen(vol_ref, path, vfsModeRead, &dir_ref) != errNone) {
	MemPtrFree(*names);
	*names = NULL;
	return 0;
    }

    cnt = 0;
    offset = 0;
    for (iterator = vfsIteratorStart; iterator != vfsIteratorStop; ) {
	info.nameP = name;
	info.nameBufLen = sizeof name;
	if (VFSDirEntryEnumerate(dir_ref, &iterator, &info) == errNone) {
	    len = StrLen(name);
	    if (cnt < max_cnt && offset + len < max_size) {
		(*names)[cnt] = (Char *) &(*names)[max_cnt] + offset;
		StrCopy((*names)[cnt], name);
		++cnt;
		offset += len + 1;
	    } else {
		break;
	    }
	} else {
	    break;
	}
    }

    VFSFileClose(dir_ref);

    if (cnt == 0) {
	MemPtrFree(*names);
	*names = NULL;
	return 0;
    }

    SysQSort(*names, cnt, sizeof(Char *), compare_names, 0);

    return cnt;
}

static char *read_file(UInt16 vol_ref, const Char *path)
{
    FileRef file_ref;
    UInt32 file_size, read_size;
    char *ptr;

    if (VFSFileOpen(vol_ref, path, vfsModeRead, &file_ref) != errNone) {
	return NULL;
    }

    if (VFSFileSize(file_ref, &file_size) != errNone) {
	VFSFileClose(file_ref);
	return NULL;
    }

    ptr = MemPtrNew(file_size);
    if (ptr == NULL) {
	VFSFileClose(file_ref);
	return NULL;
    }

    if (VFSFileRead(file_ref, file_size, ptr, &read_size) != errNone
     || read_size < file_size) {
	MemPtrFree(ptr);
	VFSFileClose(file_ref);
	return NULL;
    }

    VFSFileClose(file_ref);

    return ptr;
}

void load_vfs(cell_t *tape, offset_t tape_length)
{
    FormPtr dialog;
    UInt16 vol_ref;
    int cnt, len, i, button, selection;
    Char **names, path[LIFE_DIRECTORY_LENGTH + 1 + MAX_FILENAME];
    char *ptr;

    vol_ref = find_volume();
    if (vol_ref == vfsInvalidVolRef) {
	FrmAlert(vfs_error1_id);
	return;
    }

    cnt = enum_directory(vol_ref, LIFE_DIRECTORY, &names);
    if (cnt == 0) {
	FrmCustomAlert(vfs_error2_id, LIFE_DIRECTORY, " ", " ");
	return;
    }

    dialog = FrmInitForm(load_form_id);
#ifdef USE_HANDERA_VGA
    if (handera_vga) {
	VgaFormModify(dialog, vgaFormModify160To240);
    }
#endif
    list = GetObjectPtrIn(dialog, load_list_id);
    LstSetListChoices(list, names, cnt);
    if (prefs.vfs_top >= 0 && prefs.vfs_top < cnt) {
	LstSetTopItem(list, prefs.vfs_top);
    }
    len = StrLen(prefs.vfs_name);
    if (prefs.vfs_selection >= 0 && prefs.vfs_selection < cnt
     && StrNCompare(names[prefs.vfs_selection], prefs.vfs_name, len) == 0) {
	i = prefs.vfs_selection;
    } else {
	for (i = 0; i < cnt; ++i) {
	    if (StrNCaselessCompare(names[i], prefs.vfs_name, len) >= 0) {
		break;
	    }
	}
	if (i == cnt) {
	    i = 0;
	}
    }
    LstSetSelection(list, i);
    LstMakeItemVisible(list, i);
    FrmSetEventHandler(dialog, handle_load_dialog_event);

    button = FrmDoDialog(dialog);
    selection = LstGetSelection(list);
    prefs.vfs_top = GetTopItem(list);
    prefs.vfs_selection = selection;
    if (button == load_ok_id && selection != noListSelection) {
	StrPrintF(path, "%s/%s", LIFE_DIRECTORY, names[selection]);
	StrNCopy(prefs.vfs_name, names[selection], VFS_NAME_LENGTH - 1);
	prefs.vfs_name[VFS_NAME_LENGTH - 1] = '\0';
    }
    FrmDeleteForm(dialog);
    MemPtrFree(names);

    if (button == load_ok_id && selection != noListSelection) {
	ptr = read_file(vol_ref, path);
	if (ptr == NULL) {
	    FrmCustomAlert(vfs_error3_id,
		    path + LIFE_DIRECTORY_LENGTH + 1, " ", " ");
	} else {
	    offset_t length;

	    length = load_life(ptr, MemPtrSize(ptr), NULL);

	    if (length <= 0) {
		FrmAlert(load_error4_id);
	    } else if (length >= tape_length / 2 - 1) {
		FrmAlert(load_error5_id);
	    } else {
		length = load_life(ptr, MemPtrSize(ptr), tape);
		reset_life(length);
	    }

	    MemPtrFree(ptr);
	}
    }
}
